#
#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#   #
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#   #
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#   #
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)


import json
import logging
import logging.config
import pathlib

from PySide6.QtCore import QThreadPool
from PySide6.QtGui import QAction, QFont, QIcon
from PySide6.QtWidgets import (QApplication, QDialog, QMainWindow, QMessageBox,
                               QVBoxLayout, QWidget)

import revedaEditor.backend.hdlBackEnd as hdl
import revedaEditor.backend.importViews as imv
import revedaEditor.gui.editorWindows as edw
import revedaEditor.gui.fileDialogues as fd
import revedaEditor.gui.pythonConsole as pcon
import revinit


class mainwContainer(QWidget):
    """
    Definition for the main app window layout.
    """

    def __init__(self, parent):
        super().__init__(parent=parent)
        self.parent = parent
        self.console = pcon.pythonConsole(globals())
        self.init_UI()

    def init_UI(self):
        # treeView = designLibrariesView(self)
        self.console.setfont(QFont("Fira Mono Regular", 12))
        self.console.writeoutput(
            f"Welcome to Revolution EDA version" f" {revinit.__version__}")
        self.console.writeoutput("Revolution Semiconductor (C) 2023.")
        # layout statements, using a grid layout
        gLayout = QVBoxLayout()
        gLayout.setSpacing(10)
        gLayout.addWidget(self.console)
        self.setLayout(gLayout)


#
class MainWindow(QMainWindow):
    def __init__(self):
        super().__init__()
        self.resize(900, 300)
        self._createActions()
        self._createMenuBar()
        self._createTriggers()
        self.cellViews = ["schematic", "symbol", "layout", "veriloga", "config",
            "spice", ]
        self.switchViewList = ["schematic", "veriloga", "spice", "symbol"]
        self.stopViewList = ["symbol"]
        self.simulationPath = pathlib.Path.cwd().parent
        self.openViews = dict()
        # create container to position all widgets
        self.centralW = mainwContainer(self)
        self.setCentralWidget(self.centralW)
        self.mainW_statusbar = self.statusBar()
        self.mainW_statusbar.showMessage("Ready")
        self.app = QApplication.instance()
        # logger name is 'reveda'
        self.logger = logging.getLogger("reveda")
        # library definition file path
        self.runPath = pathlib.Path.cwd()
        # look for library.json file where the script is invoked
        self.libraryPathObj = self.runPath.joinpath("library.json")
        self.libraryDict = self.readLibDefFile(self.libraryPathObj)
        self.libraryBrowser = edw.libraryBrowser(self)
        self.logger_def()
        # revEDAPathObj = Path(__file__)
        # library definition file path
        self.runPath = pathlib.Path.cwd()
        # look for library.json file where the script is invoked
        self.libraryPathObj = self.runPath.joinpath("library.json")
        self.libraryDict = self.readLibDefFile(self.libraryPathObj)
        self.textEditorPath = str(self.runPath)
        self.threadPool = QThreadPool.globalInstance()
        self.confFilePath = self.runPath.joinpath("reveda.conf")
        self.loadState()

    def _createMenuBar(self):
        self.mainW_menubar = self.menuBar()
        self.mainW_menubar.setNativeMenuBar(False)
        # Returns QMenu object.
        self.menuFile = self.mainW_menubar.addMenu("&File")
        self.menuTools = self.mainW_menubar.addMenu("&Tools")
        self.importTools = self.menuTools.addMenu("&Import")
        self.menuOptions = self.mainW_menubar.addMenu("&Options")
        self.menuHelp = self.mainW_menubar.addMenu("&Help")
        self.menuFile.addAction(self.exitAction)
        self.menuTools.addAction(self.libraryBrowserAction)
        self.importTools.addAction(self.importVerilogaAction)
        self.menuOptions.addAction(self.optionsAction)
        # self.menuHelp.addAction(self.helpAction)
        # self.menuHelp.addAction(self.aboutAction)
        self.menuOptions.addAction(self.optionsAction)

    def _createActions(self):
        exitIcon = QIcon(":/icons/external.png")
        self.exitAction = QAction(exitIcon, "Exit", self)
        self.exitAction.setShortcut("Ctrl+Q")
        importVerilogaIcon = QIcon(":/icons/document-import.png")
        self.importVerilogaAction = QAction(importVerilogaIcon,
            "Import Verilog-a file...")
        openLibIcon = QIcon(":/icons/database--pencil.png")
        self.libraryBrowserAction = QAction(openLibIcon, "Library Browser", self)
        optionsIcon = QIcon(":/icons/resource-monitor.png")
        self.optionsAction = QAction(optionsIcon, "Options...", self)

    def _createTriggers(self):
        self.exitAction.triggered.connect(self.exitApp)  # type: ignore
        self.libraryBrowserAction.triggered.connect(self.libraryBrowserClick)
        self.importVerilogaAction.triggered.connect(self.importVerilogaClick)
        self.optionsAction.triggered.connect(self.optionsClick)

    def readLibDefFile(self, libPath: pathlib.Path):
        libraryDict = dict()
        data = dict()
        if libPath.exists():
            with libPath.open(mode="r") as f:
                data = json.load(f)
            if data.get("libdefs") is not None:
                for key, value in data["libdefs"].items():
                    libraryDict[key] = pathlib.Path(value)
            elif data.get("include") is not None:
                for item in data.get("include"):
                    libraryDict.update(self.readLibDefFile(pathlib.Path(item)))
        return libraryDict

    # open library browser window
    def libraryBrowserClick(self):
        self.libraryBrowser.show()
        self.libraryBrowser.raise_()

    def logger_def(self):
        logging.basicConfig(level=logging.INFO)

        c_handler = logging.StreamHandler(stream=self.centralW.console)
        c_handler.setLevel(logging.INFO)
        c_format = logging.Formatter("%(levelname)s - %(message)s")
        c_handler.setFormatter(c_format)
        f_handler = logging.FileHandler("reveda.log")
        f_handler.setLevel(logging.INFO)
        f_format = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(message)s")
        f_handler.setFormatter(f_format)
        self.logger.addHandler(c_handler)
        self.logger.addHandler(f_handler)

    def closeEvent(self, event):
        reply = QMessageBox.question(self, 'Confirm Exit',
                                     'Are you sure you want to exit?',
                                     QMessageBox.Yes | QMessageBox.No,
                                     QMessageBox.No)
        if reply == QMessageBox.Yes:
            self.app.closeAllWindows()
        else:
            event.ignore()

    def optionsClick(self):
        dlg = fd.appProperties(self)
        dlg.editorPathEdit.setText(self.textEditorPath)
        dlg.simPathEdit.setText(str(self.simulationPath))
        dlg.switchViewsEdit.setText(", ".join(self.switchViewList))
        dlg.stopViewsEdit.setText(", ".join(self.stopViewList))

        if dlg.exec() == QDialog.Accepted:
            self.textEditorPath = dlg.editorPathEdit.text()
            self.simulationPath = pathlib.Path(dlg.simPathEdit.text())
            self.switchViewList = [switchView.strip() for switchView in
                dlg.switchViewsEdit.text().split(",")]
            self.stopViewList = [stopView.strip() for stopView in
                dlg.stopViewsEdit.text().split(",")]
            if dlg.optionSaveBox.isChecked():
                self.saveState()

    def importVerilogaClick(self):
        """
        Import a verilog-a view and add it to a design library
        """
        libraryModel = self.libraryBrowser.libraryModel
        importDlg = fd.importVerilogaCellDialogue(libraryModel, self)
        importDlg.vaViewName.setText("veriloga")
        if importDlg.exec() == QDialog.Accepted:
            importedVaObj = hdl.verilogaC(
                pathlib.Path(importDlg.vaFileEdit.text()))
            vaViewItemTuple = imv.createVaView(self, importDlg, libraryModel,
                                               importedVaObj)
            if importDlg.symbolCheckBox.isChecked():
                imv.createVaSymbol(self, vaViewItemTuple, self.libraryDict,
                                   self.libraryBrowser, importedVaObj)

    def loadState(self):

        if self.confFilePath.exists():
            self.logger.info(f'Configuration file: {self.confFilePath} exists')
            with self.confFilePath.open(mode="r") as f:
                items = json.load(f)
            self.textEditorPath = items.get("textEditorPath")
            self.simulationPath = pathlib.Path(items.get("simulationPath"))
            if items.get("switchViewList")[0] != '':
                self.switchViewList = items.get("switchViewList")
            if items.get("stopViewList")[0] != '':
                self.stopViewList = items.get("stopViewList")

    def saveState(self):
        items = {"textEditorPath": self.textEditorPath,
            "simulationPath": str(self.simulationPath),
            "switchViewList": self.switchViewList,
            "stopViewList": self.stopViewList,

        }
        with self.confFilePath.open(mode="w", encoding="utf") as f:
            json.dump(items, f, indent=4)

    def exitApp(self):
        self.app.closeAllWindows()



#
#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#   #
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#   #
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#   #
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)

from collections import namedtuple

viewTuple = namedtuple('viewTuple', ['libraryName', 'cellName', 'viewName'])
cellTuple = namedtuple('cellTuple', ['libraryName', 'cellName'])
viewItemTuple = namedtuple('viewItemTuple', ['libraryItem', 'cellItem',
                                             'viewItem'])
pinNetTuple = namedtuple('pinNetTuple',  ['pin', 'net', 'start'])
netTuple = namedtuple('netTuple', ['net', 'start'])
# this namedtuple is used to collect information on dashed lines.
# net is the dashed line
# index is the index of the self (schematicNet) where dashed line ends,
# 0: start,
# 1: end
# orient is True if self is horizontal, otherwise False
netEndTuple = namedtuple('netEndTuple', ['net', 'index', 'orient'])




#   “Commons Clause” License Condition v1.0
#  #
#   The Software is provided to you by the Licensor under the License, as defined
#   below, subject to the following condition.
#  #
#   Without limiting other conditions in the License, the grant of rights under the
#   License will not include, and the License does not grant to you, the right to
#   Sell the Software.
#  #
#   For purposes of the foregoing, “Sell” means practicing any or all of the rights
#   granted to you under the License to provide to third parties, for a fee or other
#   consideration (including without limitation fees for hosting or consulting/
#   support services related to the Software), a product or service whose value
#   derives, entirely or substantially, from the functionality of the Software. Any
#   license notice or attribution required by the License must also include this
#   Commons Clause License Condition notice.
#  #
#   Software: Revolution EDA
#   License: Mozilla Public License 2.0
#   Licensor: Revolution Semiconductor (Registered in the Netherlands)

# Load symbol and maybe later schematic from json file.
# import pathlib

import json

from PySide6.QtCore import (QPoint, )  # QtCore

import revedaEditor.common.net as net
import revedaEditor.common.pens as pens
import revedaEditor.common.shape as shp
import revedaEditor.fileio.symbolEncoder as se


def createSymbolItems(item, gridTuple):
    """
    Create symbol items from json file.
    """
    if item["type"] == "rect":
        return createRectItem(item, gridTuple)
    elif item["type"] == "circle":
        return createCircleItem(item, gridTuple)
    elif item["type"] == "arc":
        return createArcItem(item, gridTuple)
    elif item["type"] == "line":
        return createLineItem(item, gridTuple)
    elif item["type"] == "pin":
        return createPinItem(item, gridTuple)
    elif item["type"] == "label":
        return createLabelItem(item, gridTuple)


def createRectItem(item, gridTuple):
    """
    Create symbol items from json file.
    """
    start = QPoint(item["rect"][0], item["rect"][1])
    end = QPoint(item["rect"][2], item["rect"][3])
    pen = pens.sPen.returnPen(item['pen'])
    rect = shp.rectangle(start, end, pen,
                         gridTuple)  # note that we are using grid values for scene
    rect.setPos(QPoint(item["loc"][0], item["loc"][1]), )
    rect.angle = item["ang"]
    return rect


def createCircleItem(item, gridTuple):
    centre = QPoint(item["cen"][0], item["cen"][1])
    end = QPoint(item["end"][0], item["end"][1])
    pen = pens.sPen.returnPen(item['pen'])
    circle = shp.circle(centre, end, pen,
                        gridTuple)  # note that we are using grid values for scene
    circle.setPos(QPoint(item["loc"][0], item["loc"][1]), )
    circle.angle = item["ang"]
    return circle


def createArcItem(item, gridTuple):
    start = QPoint(item["st"][0], item["st"][1])
    end = QPoint(item["end"][0], item["end"][1])
    pen = pens.sPen.returnPen(item['pen'])
    arc = shp.arc(start, end, pen,
                  gridTuple)  # note that we are using grid values for scene
    arc.setPos(QPoint(item["loc"][0], item["loc"][1]))
    arc.angle = item["ang"]
    return arc


def createLineItem(item, gridTuple):
    start = QPoint(item["st"][0], item["st"][1])
    end = QPoint(item["end"][0], item["end"][1])
    pen = pens.sPen.returnPen(item['pen'])
    line = shp.line(start, end, pen, gridTuple)
    line.setPos(QPoint(item["loc"][0], item["loc"][1]))
    line.angle = item["ang"]
    return line


def createPinItem(item, gridTuple):
    start = QPoint(item["st"][0], item["st"][1])
    pen = pens.sPen.returnPen(item['pen'])
    pin = shp.pin(start, pen, item["nam"], item["pd"], item["pt"], gridTuple, )
    pin.setPos(QPoint(item["loc"][0], item["loc"][1]))
    pin.angle = item["ang"]
    return pin


def createLabelItem(item, gridTuple):
    start = QPoint(item["st"][0], item["st"][1])
    pen = pens.sPen.returnPen(item['pen'])
    label = shp.label(start, pen, item["def"], gridTuple, item["lt"], item["ht"],
                      item["al"], item["or"], item["use"], )
    label.setPos(QPoint(item["loc"][0], item["loc"][1]))
    label.angle = item["ang"]
    label.labelName = item["nam"]
    label.labelText = item["txt"]
    label.labelVisible = item["vis"]
    label.labelValue = item["val"]
    return label


def createTextItem(item, gridTuple: (int, int)):
    start = QPoint(item["st"][0], item["st"][1])
    pen = pens.sPen.returnPen(item['pen'])
    text = shp.text(start, pen, item['tc'], gridTuple, item['ff'], item['fs'],
                    item['th'], item['ta'], item['to'])
    text.setPos(QPoint(item["loc"][0], item["loc"][1]))
    return text


def createSymbolAttribute(item):
    if item["type"] == "attr":
        return se.symbolAttribute(item["nam"], item["def"])


def createSchematicItems(item, libraryDict, viewName: str, gridTuple: (int, int)):
    """
    Create schematic items from json file.
    """
    if item["type"] == "symbolShape":
        libraryPath = libraryDict.get(item["lib"])
        if libraryPath is None:
            print(f'{item["lib"]} cannot be found.')
            return None
        cell = item["cell"]
        instCounter = item["ic"]
        itemShapes = list()
        symbolAttributes = dict()
        labelDict = item["ld"]
        draftPen = pens.sPen.returnPen('draftPen')
        # find the symbol file
        file = libraryPath.joinpath(cell, viewName + ".json")
        # load json file and create shapes
        with open(file, "r") as temp:
            try:
                shapes = json.load(temp)
                for shape in shapes[1:]:
                    if shape["type"] == "rect":
                        itemShapes.append(createRectItem(shape, gridTuple))
                    elif shape["type"] == "circle":
                        itemShapes.append(createCircleItem(shape, gridTuple))
                    elif shape["type"] == "arc":
                        itemShapes.append(createArcItem(shape, gridTuple))
                    elif shape["type"] == "line":
                        itemShapes.append(createLineItem(shape, gridTuple))
                    elif shape["type"] == "pin":
                        itemShapes.append(createPinItem(shape, gridTuple))
                    elif shape["type"] == "label":
                        itemShapes.append(createLabelItem(shape, gridTuple))
                    # just recreate attributes dictionary
                    elif shape["type"] == "attr":
                        symbolAttributes[shape["nam"]] = shape["def"]
            except json.decoder.JSONDecodeError:
                print("Error: Invalid Symbol file")
        symbolInstance = shp.symbolShape(draftPen, gridTuple, itemShapes,
                                         symbolAttributes)
        symbolInstance.libraryName = item["lib"]
        symbolInstance.cellName = item["cell"]
        symbolInstance.counter = instCounter
        symbolInstance.instanceName = item["nam"]
        symbolInstance.angle = item["ang"]
        symbolInstance.viewName = viewName
        symbolInstance.attributes = symbolAttributes
        for labelItem in symbolInstance.labels.values():
            if labelItem.labelName in labelDict.keys():
                labelItem.labelValue = labelDict[labelItem.labelName][0]
                labelItem.labelVisible = labelDict[labelItem.labelName][1]
        symbolInstance.setPos(item["loc"][0], item["loc"][1])
        return symbolInstance


def createSchematicNets(item):
    """
    Create schematic items from json file.
    """
    if item["type"] == "schematicNet":
        start = QPoint(item["st"][0], item["st"][1])
        end = QPoint(item["end"][0], item["end"][1])
        position = QPoint(item["loc"][0], item["loc"][1])
        pen = pens.sPen.returnPen(item['pen'])
        netItem = net.schematicNet(start, end, pen)
        netItem.name = item["nam"]
        netItem.nameSet = item["ns"]
        netItem.setPos(position)
        return netItem


def createSchematicPins(item, gridTuple):
    """
    Create schematic items from json file.
    """
    if item["type"] == "schematicPin":
        start = QPoint(item["st"][0], item["st"][1])
        pen = pens.sPen.returnPen(item['pen'])
        pinName = item["pn"]
        pinDir = item["pd"]
        pinType = item["pt"]
        pinItem = shp.schematicPin(start, pen, pinName, pinDir, pinType,
                                   gridTuple)
        pinItem.setPos(QPoint(item["loc"][0], item["loc"][1]))
        pinItem.angle = item["ang"]
        return pinItem



#  “Commons Clause” License Condition v1.0
#
#  The Software is provided to you by the Licensor under the License, as defined
#  below, subject to the following condition.
#
#  Without limiting other conditions in the License, the grant of rights under the
#  License will not include, and the License does not grant to you, the right to
#  Sell the Software.
#
#  For purposes of the foregoing, “Sell” means practicing any or all of the rights
#  granted to you under the License to provide to third parties, for a fee or other
#  consideration (including without limitation fees for hosting or consulting/
#  support services related to the Software), a product or service whose value
#  derives, entirely or substantially, from the functionality of the Software. Any
#  license notice or attribution required by the License must also include this
#  Commons Clause License Condition notice.
#
#  Software: Revolution EDA
#  License: Mozilla Public License 2.0
#  Licensor: Revolution Semiconductor (Registered in the Netherlands)
#


# version
__version__ = "0.5.1"



#   “Commons Clause” License Condition v1.0
#  #
#   The Software is provided to you by the Licensor under the License, as defined
#   below, subject to the following condition.
#  #
#   Without limiting other conditions in the License, the grant of rights under the
#   License will not include, and the License does not grant to you, the right to
#   Sell the Software.
#  #
#   For purposes of the foregoing, “Sell” means practicing any or all of the rights
#   granted to you under the License to provide to third parties, for a fee or other
#   consideration (including without limitation fees for hosting or consulting/
#   support services related to the Software), a product or service whose value
#   derives, entirely or substantially, from the functionality of the Software. Any
#   license notice or attribution required by the License must also include this
#   Commons Clause License Condition notice.
#  #
#   Software: Revolution EDA
#   License: Mozilla Public License 2.0
#   Licensor: Revolution Semiconductor (Registered in the Netherlands)

import os
import sys
from PySide6.QtWidgets import (QApplication)
from PySide6.QtGui import QScreen, QGuiApplication
import revedaEditor.gui.revedaMain as rvm
import revedaEditor.gui.pythonConsole as pcon
from contextlib import redirect_stdout, redirect_stderr
from dotenv import load_dotenv
from pathlib import Path


# simulation window

class revedaApp(QApplication):
    """
    Initializes the class instance and sets the paths to the revedaeditor and revedasim if the corresponding 
    environment variables exist. Also adds the paths to the system path if available.

    Returns:
        None
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # Load environment variables
        load_dotenv()
        reveda_runpathObj = Path(__file__).resolve().parent
        self.revedaeditor_path = None
        self.revedasim_path = None
        self.reveda_pdk_path = None
        self.setPaths(reveda_runpathObj)

    def setPaths(self, reveda_runpathObj):
        # Set the paths to the revedaeditor and revedasim
        self.revedaeditor_path = os.environ.get("REVEDAEDIT_PATH", None)
        if self.revedaeditor_path:
            if Path(self.revedaeditor_path).is_absolute():
                self.revedaeditor_pathObj = Path(self.revedaeditor_path)
            else:
                self.revedaeditor_pathObj = reveda_runpathObj.joinpath(
                    self.revedaeditor_path)
        self.revedasim_path = os.environ.get("REVEDASIM_PATH", None)
        if self.revedasim_path:
            if Path(self.revedasim_path).is_absolute():
                self.revedasim_pathObj = Path(self.revedasim_path)
            else:
                self.revedasim_pathObj = reveda_runpathObj.joinpath(self.revedasim_path)
            sys.path.append(self.revedasim_pathObj)
        self.reveda_pdk_path = os.environ.get("REVEDA_PDK_PATH", None)
        if self.reveda_pdk_path:
            if Path(self.reveda_pdk_path).is_absolute():
                self.reveda_pdk_pathObj = Path(self.reveda_pdk_path)
            else:
                self.reveda_pdk_pathObj = reveda_runpathObj.joinpath(self.reveda_pdk_path)
            sys.path.append(self.reveda_pdk_pathObj)


def main():
    # Start Main application window
    app = revedaApp(sys.argv)
    app.setStyle("Fusion")

    mainW = rvm.MainWindow()
    mainW.setWindowTitle("Revolution EDA")

    # empty argument as there is no parent window.
    redirect = pcon.Redirect(mainW.centralW.console.errorwrite)
    with redirect_stdout(mainW.centralW.console), redirect_stderr(redirect):
        mainW.show()
        sys.exit(app.exec())


if __name__ == "__main__":
    main()





#   “Commons Clause” License Condition v1.0
#  #
#   The Software is provided to you by the Licensor under the License, as defined
#   below, subject to the following condition.
#  #
#   Without limiting other conditions in the License, the grant of rights under the
#   License will not include, and the License does not grant to you, the right to
#   Sell the Software.
#  #
#   For purposes of the foregoing, “Sell” means practicing any or all of the rights
#   granted to you under the License to provide to third parties, for a fee or other
#   consideration (including without limitation fees for hosting or consulting/
#   support services related to the Software), a product or service whose value
#   derives, entirely or substantially, from the functionality of the Software. Any
#   license notice or attribution required by the License must also include this
#   Commons Clause License Condition notice.
#  #
#   Software: Revolution EDA
#   License: Mozilla Public License 2.0
#   Licensor: Revolution Semiconductor (Registered in the Netherlands)

from PySide6.QtGui import (QFont, QFontMetrics, QFontDatabase)

class font(QFont):
    def __init__(self,fontFamily:str, fontStyle:str, fontSize:int, kerning:bool,logger):
        self._fontFamily = fontFamily
        self._fontStyle = fontStyle
        self._fontSize = fontSize
        self._kerning = kerning
        self._logger = logger
        self.setFamily(self._fontFamily)
        self.setStyle(self._fontStyle)
        self.setPointSize(self._fontSize)
        self.setKerning(self._kerning)
        self.setStyleHint(QFont.TypeWriter)

    @property
    def fontFamily(self):
        return self._fontFamily

    @fontFamily.setter
    def fontFamily(self, value:str):
        fontFamilies = QFontDatabase.families(QFontDatabase.Latin)
        fixedFamilies = [family for family in fontFamilies if
                         QFontDatabase.isFixedPitch(family)]
        if value not in fixedFamilies:
            self._logger.warning('No such font family present in the system. Another '
                                 'font '
                           'will be substituted.')
        self._fontFamily = value
        self.setFamily(self._fontFamily)

    @property
    def fontStyle(self):
        return self._fontStyle

    @fontStyle.setter
    def fontStyle(self,value:str):
        fstyles = QFontDatabase.styles(self._fontFamily)
        if value not in fstyles:
            self._logger.warning(f'No matching style is available. Available styles '
                                 f'are: {",".join(fstyles)}')
        else:
            self.setStyle(value)





#
#    “Commons Clause” License Condition v1.0
#   #
#    The Software is provided to you by the Licensor under the License, as defined
#    below, subject to the following condition.
#   #
#    Without limiting other conditions in the License, the grant of rights under the
#    License will not include, and the License does not grant to you, the right to
#    Sell the Software.
#   #
#    For purposes of the foregoing, “Sell” means practicing any or all of the rights
#    granted to you under the License to provide to third parties, for a fee or other
#    consideration (including without limitation fees for hosting or consulting/
#    support services related to the Software), a product or service whose value
#    derives, entirely or substantially, from the functionality of the Software. Any
#    license notice or attribution required by the License must also include this
#    Commons Clause License Condition notice.
#   #
#    Software: Revolution EDA
#    License: Mozilla Public License 2.0
#    Licensor: Revolution Semiconductor (Registered in the Netherlands)

import json
import pathlib
import shutil

from PySide6.QtCore import (Qt, QPoint)
from PySide6.QtWidgets import (QMainWindow, QDialog)

import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.hdlBackEnd as hdl
import revedaEditor.backend.libraryMethods as libm  # library view functions
import revedaEditor.backend.schBackEnd as scb  # import the backend
import revedaEditor.common.pens as pens
import revedaEditor.common.shape as shp
import revedaEditor.fileio.symbolEncoder as se
import revedaEditor.gui.editorWindows as edw
import revedaEditor.gui.propertyDialogues as pdlg


def createVaView(parent: QMainWindow, importDlg: QDialog, libraryModel:
edw.designLibrariesModel, importedVaObj: hdl.verilogaC):
    importedVaFilePathObj = pathlib.Path(importDlg.vaFileEdit.text())
    libItem = libm.getLibItem(
        libraryModel, importDlg.libNamesCB.currentText()
    )
    libItemRow = libItem.row()
    libCellNames = [
        libraryModel.item(libItemRow).child(i).cellName
        for i in range(libraryModel.item(libItemRow).rowCount())
    ]
    cellName = importDlg.cellNamesCB.currentText().strip()
    if cellName not in libCellNames and cellName != "":
        scb.createCell(parent, libraryModel, libItem, cellName)
    cellItem = libm.getCellItem(libItem, cellName)
    newVaFilePathObj = cellItem.data(Qt.UserRole + 2).joinpath(
        importedVaFilePathObj.name)
    vaItem = scb.createCellView(parent, importDlg.vaViewName.text(), cellItem)
    tempFilePathObj = importedVaFilePathObj.with_suffix(".temp")
    shutil.copy(importedVaFilePathObj, tempFilePathObj)

    shutil.copy(tempFilePathObj, newVaFilePathObj)
    tempFilePathObj.unlink()
    items = list()
    items.insert(0, {"cellView": "veriloga"})
    items.insert(1, {"filePath": str(newVaFilePathObj.name)})
    items.insert(2, {"vaModule": importedVaObj.vaModule})
    items.insert(3, {"netlistLine": importedVaObj.netListLine})
    with vaItem.data(Qt.UserRole + 2).open(mode="w") as f:
        json.dump(items, f, indent=4)
    return ddef.viewItemTuple(libItem, cellItem, vaItem)


def createVaSymbol(parent: QMainWindow,
                   vaItemTuple: ddef.viewItemTuple,
                   libraryDict: dict,
                   libraryBrowser: edw.libraryBrowser,
                   importedVaObj: hdl.verilogaC):
    symbolViewItem = scb.createCellView(parent, "symbol", vaItemTuple.cellItem)
    newVaFilePathObj = vaItemTuple.cellItem.data(Qt.UserRole + 2).joinpath(
        importedVaObj.pathObj.name)
    symbolWindow = edw.symbolEditor(
        symbolViewItem,
        libraryDict,
        libraryBrowser.libBrowserCont.designView,
    )
    symbolPen = pens.sPen.returnPen("symbolPen")
    labelPen = pens.sPen.returnPen("labelPen")
    pinPen = pens.sPen.returnPen(("pinPen"))
    dlg = pdlg.symbolCreateDialog(
        parent,
        importedVaObj.inPins,
        importedVaObj.outPins,
        importedVaObj.inoutPins,
    )
    dlg.leftPinsEdit.setText(",".join(importedVaObj.inPins))
    dlg.rightPinsEdit.setText(",".join(importedVaObj.outPins))
    dlg.topPinsEdit.setText(",".join(importedVaObj.inoutPins))

    if dlg.exec() == QDialog.Accepted:
        try:
            leftPinNames = list(
                filter(
                    None,
                    [
                        pinName.strip()
                        for pinName in dlg.leftPinsEdit.text().split(
                        ","
                    )
                    ],
                )
            )
            rightPinNames = list(
                filter(
                    None,
                    [
                        pinName.strip()
                        for pinName in dlg.rightPinsEdit.text().split(
                        ","
                    )
                    ],
                )
            )
            topPinNames = list(
                filter(
                    None,
                    [
                        pinName.strip()
                        for pinName in dlg.topPinsEdit.text().split(",")
                    ],
                )
            )
            bottomPinNames = list(
                filter(
                    None,
                    [
                        pinName.strip()
                        for pinName in dlg.bottomPinsEdit.text().split(
                        ","
                    )
                    ],
                )
            )
            stubLength = int(float(dlg.stubLengthEdit.text().strip()))
            pinDistance = int(float(dlg.pinDistanceEdit.text().strip()))
            rectXDim = (
                               max(len(topPinNames), len(bottomPinNames)) + 1
                       ) * pinDistance
            rectYDim = (
                               max(len(leftPinNames), len(rightPinNames)) + 1
                       ) * pinDistance
        except ValueError:
            parent.logger.error("Enter valid value")
        symbolScene = symbolWindow.centralW.scene
        symbolScene.rectDraw(
            QPoint(0, 0),
            QPoint(rectXDim, rectYDim),
            symbolPen,
            symbolScene.gridTuple,
        )
        symbolScene.labelDraw(
            QPoint(int(0.25 * rectXDim), int(0.4 * rectYDim)),
            labelPen,
            "[@cellName]",
            "NLPLabel",
            "12",
            "Center",
            "R0",
            "Instance",
            symbolScene.gridTuple,
        )
        symbolScene.labelDraw(
            QPoint(int(rectXDim), int(-0.2 * rectYDim)),
            labelPen,
            "[@instName]",
            "NLPLabel",
            "12",
            "Center",
            "R0",
            "Instance",
            symbolScene.gridTuple,
        )
        vaFileLabel = symbolScene.labelDraw(
            QPoint(int(0.25 * rectXDim), int(0.6 * rectYDim)),
            labelPen,
            f"[@vaFile:vaFile=%:vaFile={str(newVaFilePathObj)}]",
            "NLPLabel",
            "12",
            "Center",
            "R0",
            "Instance",
            symbolScene.gridTuple,
        )
        vaFileLabel.labelVisible = False
        vaModuleLabel = symbolScene.labelDraw(
            QPoint(int(0.25 * rectXDim), int(0.8 * rectYDim)),
            labelPen,
            f"[@vaModule:vaModule=%:vaModule={importedVaObj.vaModule}]",
            "NLPLabel",
            "12",
            "Center",
            "R0",
            "Instance",
            symbolScene.gridTuple,
        )
        vaModuleLabel.labelVisible = False
        vaModelLabel = symbolScene.labelDraw(
            QPoint(int(0.25 * rectXDim), int(1 * rectYDim)),
            labelPen,
            f"[@vaModel:vaModel=%:vaModel={importedVaObj.vaModule}Model]",
            "NLPLabel",
            "12",
            "Center",
            "R0",
            "Instance",
            symbolScene.gridTuple,
        )
        vaModelLabel.labelVisible = False
        i = 0
        instParamNum = len(importedVaObj.instanceParams)
        for key, value in importedVaObj.instanceParams.items():
            symbolScene.labelDraw(
                QPoint(
                    int(rectXDim),
                    int(i * 0.2 * rectYDim / instParamNum),
                ),
                labelPen,
                f"[@{key}:{key}=%:{key}={value}]",
                "NLPLabel",
                "12",
                "Center",
                "R0",
                "Instance",
                symbolScene.gridTuple,
            )

        leftPinLocs = [
            QPoint(-stubLength, (i + 1) * pinDistance)
            for i in range(len(leftPinNames))
        ]
        rightPinLocs = [
            QPoint(rectXDim + stubLength, (i + 1) * pinDistance)
            for i in range(len(rightPinNames))
        ]
        bottomPinLocs = [
            QPoint((i + 1) * pinDistance, rectYDim + stubLength)
            for i in range(len(bottomPinNames))
        ]
        topPinLocs = [
            QPoint((i + 1) * pinDistance, -stubLength)
            for i in range(len(topPinNames))
        ]
        for i, pinName in enumerate(leftPinNames):
            symbolScene.lineDraw(
                leftPinLocs[i],
                leftPinLocs[i] + QPoint(stubLength, 0),
                symbolScene.symbolPen,
                symbolScene.gridTuple,
            )
            symbolScene.addItem(
                shp.pin(leftPinLocs[i], pinPen, pinName)
            )
        for i, pinName in enumerate(rightPinNames):
            symbolScene.lineDraw(
                rightPinLocs[i],
                rightPinLocs[i] + QPoint(-stubLength, 0),
                symbolScene.symbolPen,
                symbolScene.gridTuple,
            )
            symbolScene.addItem(
                shp.pin(rightPinLocs[i], pinPen, pinName)
            )
        for i, pinName in enumerate(topPinNames):
            symbolScene.lineDraw(
                topPinLocs[i],
                topPinLocs[i] + QPoint(0, stubLength),
                symbolScene.symbolPen,
                symbolScene.gridTuple,
            )
            symbolScene.addItem(shp.pin(topPinLocs[i], pinPen, pinName))
        for i, pinName in enumerate(bottomPinNames):
            symbolScene.lineDraw(
                bottomPinLocs[i],
                bottomPinLocs[i] + QPoint(0, -stubLength),
                symbolScene.symbolPen,
                symbolScene.gridTuple,
            )
            symbolScene.addItem(
                shp.pin(bottomPinLocs[i], pinPen, pinName)
            )
        symbolScene.attributeList = list()  # empty attribute list
        for key, value in importedVaObj.modelParams.items():
            symbolScene.attributeList.append(
                se.symbolAttribute(key, value)
            )
        symbolScene.attributeList.append(
            se.symbolAttribute(
                "XyceNetlistLine", importedVaObj.netListLine
            ))
        symbolScene.attributeList.append(
            se.symbolAttribute(
                "pinOrder", importedVaObj.pinOrder
        ))
        symbolWindow.show()
        symbolViewTuple = ddef.viewTuple(vaItemTuple.libraryItem.libraryName,
                                         vaItemTuple.cellItem.cellName, 'symbol')
        symbolWindow.libraryView.openViews[symbolViewTuple] = symbolWindow



#   “Commons Clause” License Condition v1.0
#  #
#   The Software is provided to you by the Licensor under the License, as defined
#   below, subject to the following condition.
#  #
#   Without limiting other conditions in the License, the grant of rights under the
#   License will not include, and the License does not grant to you, the right to
#   Sell the Software.
#  #
#   For purposes of the foregoing, “Sell” means practicing any or all of the rights
#   granted to you under the License to provide to third parties, for a fee or other
#   consideration (including without limitation fees for hosting or consulting/
#   support services related to the Software), a product or service whose value
#   derives, entirely or substantially, from the functionality of the Software. Any
#   license notice or attribution required by the License must also include this
#   Commons Clause License Condition notice.
#  #
#   Software: Revolution EDA
#   License: Mozilla Public License 2.0
#   Licensor: Revolution Semiconductor (Registered in the Netherlands)

import json

import revedaEditor.common.net as net
import revedaEditor.common.shape as shp


class symbolAttribute(object):
    def __init__(self, name: str, definition: str):
        self._name = name
        self._definition = definition

    def __str__(self):
        return f'{self.name}: {self.definition}'

    def __repr__(self):
        return f'{self.name}:  {self.definition}'

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        assert isinstance(value, str)
        self._name = value

    @property
    def definition(self):
        return self._definition

    @definition.setter
    def definition(self, value):
        assert isinstance(value, str)
        self._definition = value


class symbolEncoder(json.JSONEncoder):
    def default(self, item):
        if isinstance(item, shp.rectangle):
            itemDict = {"type": "rect", "rect": item.rect.getCoords(),
                        "pen": item.pen.pname,
                        "loc": (item.scenePos() - item.scene().origin).toTuple(),
                        "ang": item.angle, }
            return itemDict
        elif isinstance(item, shp.line):
            itemDict = {"type": "line", "st": item.start.toTuple(),
                        "end": item.end.toTuple(), "pen": item.pen.pname,
                        "loc": (item.scenePos() - item.scene().origin).toTuple(),
                        "ang": item.angle, }
            return itemDict
        elif isinstance(item, shp.circle):
            itemDict = {"type": "circle", "cen": item.centre.toTuple(),
                        "end": item.end.toTuple(), "pen": item.pen.pname,
                        "loc": (item.scenePos() - item.scene().origin).toTuple(),
                        "ang": item.angle, }
            return itemDict
        elif isinstance(item, shp.arc):
            itemDict = {"type": "arc", "st": item.start.toTuple(),
                        "end": item.end.toTuple(), "pen": item.pen.pname,
                        "loc": (item.scenePos() - item.scene().origin).toTuple(),
                        "ang": item.angle, }
            return itemDict
        elif isinstance(item, shp.pin):
            itemDict = {"type": "pin", "st": item.start.toTuple(),
                        "pen": item.pen.pname, "nam": item.pinName,
                        "pd": item.pinDir, "pt": item.pinType,
                        "loc": (item.scenePos() - item.scene().origin).toTuple(),
                        "ang": item.angle, }
            return itemDict
        elif isinstance(item, shp.text):
            itemDict = {"type": "text", "st": item.start.toTuple(),
                        "pen": item.pen.pname, 'tc': item.textContent,
                        'ff': item.fontFamily, 'fs': item.fontStyle,
                        'th': item.textHeight, 'ta': item.textAlignment,
                        'to': item.textOrient,
                        "loc": (item.scenePos() - item.scene().origin).toTuple(),
                        "ang": item.angle, }
            return itemDict
        elif isinstance(item, shp.label):
            itemDict = {"type": "label", "st": item.start.toTuple(),
                        "pen": item.pen.pname, "nam": item.labelName,
                        "def": item.labelDefinition,  # label as entered
                        "txt": item.labelText,  # shown label
                        "val": item.labelValue,  # label value
                        "vis": item.labelVisible,  # label visibility
                        "lt": item.labelType, "ht": item.labelHeight,
                        "al": item.labelAlign, "or": item.labelOrient,
                        "use": item.labelUse,
                        "loc": (item.scenePos() - item.scene().origin).toTuple(),
                        "ang": item.angle, }
            return itemDict
        elif isinstance(item, symbolAttribute):
            itemDict = {"type": "attr", "nam": item.name,
                        "def": item.definition, }
            return itemDict


class schematicEncoder(json.JSONEncoder):
    def default(self, item):
        if isinstance(item, shp.symbolShape):
            # only value and visibility be changed in the symbol instance.
            itemLabelDict = {
                item.labelName: [item.labelValue, item.labelVisible] for item in
                item.labels.values()}
            itemDict = {"type": "symbolShape", "lib": item.libraryName,
                        "cell": item.cellName, "view": item.viewName,
                        "nam": item.instanceName, "ic": item.counter,
                        "ld": itemLabelDict, "loc": (
                            item.scenePos() - item.scene().origin).toTuple(),
                        "ang": item.angle, }
            return itemDict
        elif isinstance(item, net.schematicNet):
            itemDict = {"type": "schematicNet", "st": item.start.toTuple(),
                        "end": item.end.toTuple(), "pen": item.pen.pname,
                        "loc": (item.scenePos() - item.scene().origin).toTuple(),
                        "nam": item.name, "ns": item.nameSet, }
            return itemDict
        elif isinstance(item, shp.schematicPin):
            itemDict = {"type": "schematicPin", "st": item.start.toTuple(),
                        "pen": item.pen.pname, "pn": item.pinName,
                        "pd": item.pinDir, "pt": item.pinType, "loc": (
                            item.scenePos() - item.scene().origin).toTuple(),
                        "ang": item.angle, }
            return itemDict
        elif isinstance(item, shp.text):
            itemDict = {"type": "text", "st": item.start.toTuple(),
                        "pen": item.pen.pname, 'tc': item.textContent,
                        'ff': item.fontFamily, 'fs': item.fontStyle,
                        'th': item.textHeight, 'ta': item.textAlignment,
                        'to': item.textOrient, 'loc': (
                            item.scenePos() - item.scene().origin).toTuple(),
                        'ang': item.angle, }
            return itemDict




#   “Commons Clause” License Condition v1.0
#  #
#   The Software is provided to you by the Licensor under the License, as defined
#   below, subject to the following condition.
#  #
#   Without limiting other conditions in the License, the grant of rights under the
#   License will not include, and the License does not grant to you, the right to
#   Sell the Software.
#  #
#   For purposes of the foregoing, “Sell” means practicing any or all of the rights
#   granted to you under the License to provide to third parties, for a fee or other
#   consideration (including without limitation fees for hosting or consulting/
#   support services related to the Software), a product or service whose value
#   derives, entirely or substantially, from the functionality of the Software. Any
#   license notice or attribution required by the License must also include this
#   Commons Clause License Condition notice.
#  #
#   Software: Revolution EDA
#   License: Mozilla Public License 2.0
#   Licensor: Revolution Semiconductor (Registered in the Netherlands)

# net class definition.
from PySide6.QtCore import (QPoint, Qt, QLineF, QRectF, QPointF, QRect)
from PySide6.QtGui import (QPen, QStaticText, QPainterPath, QColor, QFont,)
from PySide6.QtWidgets import (QGraphicsLineItem, QGraphicsItem,
                               QGraphicsEllipseItem, QGraphicsRectItem,
                               QGraphicsSceneMouseEvent)
# import revedaEditor.common.pens as pens
import revedaEditor.backend.dataDefinitions as ddef
import revedaEditor.backend.undoStack as us


class schematicNet(QGraphicsLineItem):
    '''
    Base schematic net class.
    '''
    uses = ["SIGNAL", "ANALOG", "CLOCK", "GROUND", "POWER", ]

    def __init__(self, start: QPoint, end: QPoint, pen: QPen):
        assert isinstance(pen, QPen)
        self._pen = pen
        self._name = None
        self._horizontal = True
        self._start = start
        self._end = end
        self._nameSet = False  # if a name has been set
        self._nameAdded = False # net name is propagated to the net
        self._nameConflict = False  # if a name conflict has been detected
        super().__init__(QLineF(self._start, self._end))
        self.setPen(self._pen)
        self.setFlag(QGraphicsItem.ItemIsMovable, True)
        self.setFlag(QGraphicsItem.ItemIsSelectable, True)
        self.setFlag(QGraphicsItem.ItemSendsGeometryChanges, True)
        self.setFlag(QGraphicsItem.ItemIsFocusable, True)
        # self._createdNets = dict()
        self._endPoints = [self._start, self._end]
        self._dots = set()
        self._dotPoints = set()
        # self._touchingNets = set()
        self._dashedLines = dict()
        self._newWires = list()

    def __repr__(self):
        return f"schematicNet(start={self.mapToScene(self._start)}, " \
               f"end={self.mapToScene(self._end)}"

    def shape(self):
        '''
        Return the shape of the net.
        '''
        gridTuple = self.scene().gridTuple
        path = QPainterPath()
        path.addRect(QRectF(self.line().p1(),self.line().p2()).normalized().adjusted(
            -0.5*gridTuple[0], -0.5*gridTuple[1], 0.5*gridTuple[0], 0.5*gridTuple[1]))
        return path

    def paint(self, painter, option, widget) -> None:
        line = self.line()
        painter.setPen(self._pen)
        if self.isSelected():
            painter.setPen(QPen(Qt.blue, 2, Qt.SolidLine))
        if self.name is not None:
            if self._nameConflict:
                painter.setPen(QPen(Qt.red, 2, Qt.SolidLine))
            # if there is name conflict, draw the line and name in red.
            textLoc = line.center()
            painter.drawStaticText(textLoc, QStaticText(self.name))
        painter.drawLine(line)



    def sceneEvent(self, event):
        try:
            if self.scene().drawWire:
                return False
            super().sceneEvent(event)
            return True
        except AttributeError:
            return False

    @property
    def start(self):
        return self._start
        # return self._start

    @start.setter
    def start(self, start: QPoint):
        self._start = start
        self.prepareGeometryChange()
        self.setLine(QLineF(start, self._end))

    @property
    def end(self):
        return self._end

    @end.setter
    def end(self, end: QPoint):
        self._end = end
        self.prepareGeometryChange()
        self.setLine(QLineF(self._start, end))

    @property
    def pen(self):
        return self._pen

    @pen.setter
    def pen(self, pen: QPen):
        self._pen = pen
        self.setPen(pen)
        self.update()

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name):
        if name != "": # net name should not be an empty string
            self._name = name
            # self.nameSet = True



    @property
    def nameSet(self) -> bool:
        return self._nameSet

    @nameSet.setter
    def nameSet(self, value: bool):
        assert isinstance(value,bool)
        self._nameSet = value

    @property
    def nameConflict(self) -> bool:
        return self._nameConflict

    @nameConflict.setter
    def nameConflict(self, value: bool):
        assert isinstance(value,bool)
        self._nameConflict = value

    @property
    def length(self):
        return self.line().length()

    @property
    def endPoints(self):
        return [self.mapToScene(self._start).toPoint(), self.mapToScene(self._end).toPoint()]
        # return [self.mapToScene(self.line().p1()),self.mapToScene(self.line().p2())]


    @property
    def horizontal(self):
        self._horizontal = (self.line().p1().y() == self.line().p2().y())
        return self._horizontal

    @property
    def dots(self):
        return self._dots

    @property
    def dotPoints(self):
        return self._dotPoints

    def findDotPoints(self):
        try:
            [self.scene().removeItem(dot) for dot in self._dots]
            self._dots.clear()
            self._dotPoints.clear()
            # self._touchingNets.clear()

            if self.horizontal:
                nets = {netItem for netItem in self.scene().items(
                    self.sceneBoundingRect()) if
                        (isinstance(netItem, schematicNet) and not
                         netItem.horizontal)}

            else:
                nets = {netItem for netItem in self.scene().items(
                    self.sceneBoundingRect()) if
                        (isinstance(netItem, schematicNet) and
                         netItem.horizontal)}
            for netItem in nets:
                for selfEnd in self.endPoints:
                    if (netItem.sceneBoundingRect().contains(
                            selfEnd) and selfEnd not in
                            netItem.endPoints):
                        self._dotPoints.add(self.mapFromScene(selfEnd).toPoint())
                        # self._touchingNets.add(netItem)

            [self._dots.add(crossingDot(dotPoint,3,self.scene().wirePen)) for dotPoint in
             self._dotPoints]
            [dot.setParentItem(self) for dot in self._dots]
            [self.scene().addItem(dot) for dot in self._dots]
            # [netItem.findDotPoints() for netItem in self._touchingNets]

        except Exception as e:
            self.scene().logger.error(f'Error in net.findDotPoints: {e}')

    def createDashLines(self):
        self._dashedLines = list()
        try:
            if self.horizontal:
                nets = {netItem for netItem in self.scene().items(
                    self.sceneBoundingRect()) if
                        (isinstance(netItem, schematicNet) and not
                        netItem.horizontal)}

            else:
                nets = {netItem for netItem in self.scene().items(
                    self.sceneBoundingRect()) if
                        (isinstance(netItem, schematicNet) and
                         netItem.horizontal)}
            for netItem in nets:
                for selfEnd in self.endPoints:
                    if selfEnd in netItem.endPoints:
                        # print(f'self end:{selfEnd}, net end:'
                        #       f'{netItem.endPoints.index(selfEnd)}'
                        netItemEnd = netItem.endPoints[netItem.endPoints.index(selfEnd)]
                        self._dashedLines.append(ddef.netEndTuple(schematicNet(
                            netItemEnd, selfEnd, self.scene().otherPen),
                                        self.endPoints.index(selfEnd), self.horizontal))

        except Exception as e:
            self.scene().logger.error(f'Error in net.createDashLines: {e}')

    def extendDashLines(self):

        try:
            for netEndTuple in self._dashedLines:
                netEndTuple.net.end = self.endPoints[netEndTuple.index]

                if not netEndTuple.net.scene():
                    self.scene().addItem(netEndTuple.net)

        except Exception as e:
            self.scene().logger.error(f'Error in net.extendDashLines: {e}')

    def extendNets(self):
        try:
            for netEndTuple in self._dashedLines:
                start = netEndTuple.net.start
                end = netEndTuple.net.end
                lines = list()
                if start != end:
                    if start.y() != end.y() or start.x() != end.x():
                        if netEndTuple.orient:  # if self is horizontal
                            # | -  Start vertical, connect horizontal
                            lines.append(schematicNet(start, QPoint(end.x(), start.y()),
                                                      self.scene().wirePen))
                            lines.append(schematicNet(QPoint(end.x(), start.y()),end,
                                                      self.scene().wirePen))
                        else: # if self is vertical
                            # - |
                            lines.append(schematicNet(start, QPoint(start.x(), end.y()),
                                                      self.scene().wirePen))
                            lines.append(schematicNet(QPoint(start.x(), end.y()),end,
                                                      self.scene().wirePen))
                    elif start.y() == end.y() or start.x() == end.x():
                        lines.append(schematicNet(start, end, self.scene().wirePen))

                [self.scene().addItem(line) for line in lines]
        except Exception as e:
            self.scene().logger.error(f'Error in net.extendNets: {e}')

    def itemChange(self, change, value):

        if change == QGraphicsItem.ItemPositionChange and self.scene():
            newPos = value.toPoint()
            sceneRect = self.scene().sceneRect()
            viewRect = self.scene().views()[0].viewport().rect()

            # Keep the item inside the view rect.
            if not sceneRect.contains(newPos):
                # Keep the item inside the scene rect.
                if newPos.x() > sceneRect.right():
                    sceneRect.setRight(newPos.x())
                    viewRect.setRight(newPos.x())
                elif newPos.x() < sceneRect.left():
                    sceneRect.setLeft(newPos.x())
                    viewRect.setLeft(newPos.x())
                if newPos.y() > sceneRect.bottom():
                    sceneRect.setBottom(newPos.y())
                    viewRect.setBottom(newPos.y())
                elif newPos.y() < sceneRect.top():
                    sceneRect.setTop(newPos.y())
                    viewRect.setTop(newPos.y())
            return newPos
        elif change == QGraphicsItem.ItemPositionHasChanged and self.scene():
            # self.mergeNets()
            self.extendDashLines()
            self.findDotPoints()
        elif change == QGraphicsItem.ItemSelectedHasChanged and self.scene():
            if value:
                self.scene().editorWindow.messageLine.setText("Selected Net")
                self.mergeNets()
                self.findDotPoints()
            else:
                # self.removeDashedLines()
                # self.mergeNets()
                self.scene().editorWindow.messageLine.setText("Unselected Net")
        return super().itemChange(change, value)

    def removeDashedLines(self):
        for netEndTuple in self._dashedLines:
            self.scene().removeItem(netEndTuple.net)
            del netEndTuple

    def contextMenuEvent(self, event):
        self.scene().itemContextMenu.exec_(event.screenPos())

    def mousePressEvent(self, event):
        super().mousePressEvent(event)
        if event.button() == Qt.LeftButton:
            self.createDashLines()

    # def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent) -> None:
    #     super().mouseMoveEvent(event)
    #     # self.extendDashLines()

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent) -> None:
        super().mouseReleaseEvent(event)
        try:
            self.extendNets()
            self.removeDashedLines()
            self.mergeNets()
            self.findDotPoints()
        except Exception as e:
            self.scene().logger.error(f'Error in net.mouseReleaseEvent: {e}')


    def mergeNets(self):
        if self.horizontal:
            nets = {netItem for netItem in self.scene().items(
                self.sceneBoundingRect()) if (isinstance(netItem,schematicNet) and
                                        netItem.horizontal and netItem.pen is
                                              not self.scene().otherPen)}
            startXList = [netItem.start.x() for netItem in nets]
            endXList = [netItem.end.x() for netItem in nets]
            startXList.extend(endXList)
            startX = min(startXList)
            endX = max(startXList)
            # self.setLine(QLineF(startX, self.start.y(), endX, self.end.y()))
            self.start = QPoint(startX, self.start.y())
            self.end = QPoint(endX, self.end.y())
            for netItem in nets.difference({self}):
                self.scene().removeItem(netItem)
                del netItem
        else:
            nets = {netItem for netItem in self.scene().items(
                self.sceneBoundingRect()) if (isinstance(netItem,schematicNet)
                                and not netItem.horizontal and netItem.pen is
                                              not self.scene().otherPen)}
            startYList = [netItem.start.y() for netItem in nets]
            endYList = [netItem.end.y() for netItem in nets]
            startYList.extend(endYList)
            startY = min(startYList)
            endY = max(startYList)
            # self.setLine(QLineF(self.start.x(), startY, self.end.x(), endY))
            self.start = QPoint(self.start.x(),startY)
            self.end = QPoint(self.end.x(),endY)
            self.update()
            for netItem in nets.difference({self}):
                self.scene().removeItem(netItem)
                del netItem


class crossingDot(QGraphicsEllipseItem):
    def __init__(self, point: QPoint, radius: int, pen: QPen):
        self.radius = radius
        self._pen = pen
        self.point = point
        super().__init__(point.x() - radius, point.y() - radius, 2 * radius, 2 * radius)
        self.setPen(pen)
        self.setBrush(pen.color())

    def paint(self, painter, option, widget) -> None:
        if self.isSelected():
            painter.setPen(QPen(Qt.white, 2, Qt.SolidLine))
            painter.setBrush(Qt.white)
        else:
            painter.setPen(self._pen)
            painter.setBrush(self._pen.color())
        painter.drawEllipse(self.point, self.radius, self.radius)

class snapPointRect(QGraphicsRectItem):
    def __init__(self, centre:QPoint, width:int, pen: QPen):
        self._centre = centre
        self._width = width
        self._pen = pen
        snapRect = QRectF(QPointF(centre.x()-width*0.5, centre.y()-0.5*width), QPointF(
            centre.x()+width*0.5,centre.y()+width*0.5))

        super().__init__(snapRect)
        self.setRotation(90)

    def paint(self,painter,option,widget) -> None:
        painter.setPen(self._pen)
        painter.drawRect(self.rect)




#   “Commons Clause” License Condition v1.0
#  #
#   The Software is provided to you by the Licensor under the License, as defined
#   below, subject to the following condition.
#  #
#   Without limiting other conditions in the License, the grant of rights under the
#   License will not include, and the License does not grant to you, the right to
#   Sell the Software.
#  #
#   For purposes of the foregoing, “Sell” means practicing any or all of the rights
#   granted to you under the License to provide to third parties, for a fee or other
#   consideration (including without limitation fees for hosting or consulting/
#   support services related to the Software), a product or service whose value
#   derives, entirely or substantially, from the functionality of the Software. Any
#   license notice or attribution required by the License must also include this
#   Commons Clause License Condition notice.
#  #
#   Software: Revolution EDA
#   License: Mozilla Public License 2.0
#   Licensor: Revolution Semiconductor (Registered in the Netherlands)

import pathlib


class verilogaC:
    """
    This class represents an imported verilog-A module.
    """

    def __init__(self, pathObj: pathlib.Path):
        self._pathObj = pathObj
        keyWords = ["analog", "electrical"]
        self._vaModule = ''
        self.instanceParams = dict()
        self.modelParams = dict()
        self.inPins = list()
        self.inoutPins = list()
        self.outPins = list()
        self._netlistLine = ''
        self.statementLines = list()
        self._pinOrder = ''

        with open(self._pathObj) as f:
            self.fileLines = f.readlines()
        self.stripComments()
        self.oneLiners()
        # splitLines=[fileline.split() for fileline in fileLines]

    def stripComments(self):

        comment = False
        for line in self.fileLines:  # concatenate the lines until it reaches a ;
            stripLine = line.strip()
            if stripLine.startswith('/*'):
                comment = True
            if not comment:
                doubleSlashLoc = stripLine.find('//')
                if doubleSlashLoc > -1:
                    stripLine = stripLine[:doubleSlashLoc]
                if stripLine != '':
                    self.statementLines.append(stripLine)
            if comment and stripLine.endswith('*/'):
                comment = False

    def oneLiners(self):
        tempLine = ''
        oneLiners = list()
        for line in self.statementLines:
            stripLine = line.strip()
            if not stripLine.startswith('`include'):
                tempLine = f'{tempLine} {stripLine}'
            if stripLine.endswith(';'):
                oneLiners.append(tempLine.strip())
                splitLine = tempLine.strip().split()
                if splitLine:
                    if splitLine[0] == 'module':
                        self._vaModule = splitLine[1].split('(')[0]
                        indexLow = line.index("(")
                        indexHigh = line.index(")")
                        self.pins = [pin.strip() for pin in
                                     line[indexLow + 1: indexHigh].split(",")]
                    elif splitLine[0] == 'in' or splitLine[0] == 'input':
                        pinsList = splitLine[1].split(';')[0].split(',')
                        self.inPins.extend([pin.strip() for pin in pinsList])
                    elif splitLine[0] == 'out' or splitLine[0] == 'output':
                        pinsList = splitLine[1].split(';')[0].split(',')
                        self.outPins.extend([pin.strip() for pin in pinsList])
                    elif splitLine[0] == 'inout':
                        pinsList = splitLine[1].split(';')[0].split(',')
                        self.inoutPins.extend([pin.strip() for pin in pinsList])

                    elif splitLine[0] == "parameter":
                        paramDefPart = tempLine.split('*(')[0]
                        paramName = paramDefPart.split('=')[0].split()[-1].strip()
                        paramValue = paramDefPart.split('=')[1].split()[0].strip()

                        try:
                            paramAttr = tempLine.strip().split("(*")[1]
                        except IndexError:
                            paramAttr = ""
                        if "type" in paramAttr and '"instance"' in paramAttr:
                            # parameter value is between = and (*
                            self.instanceParams[paramName] = paramValue
                            if "xyceAlsoModel" in paramAttr and '"yes"' in paramAttr:
                                self.modelParams[paramName] = paramValue
                        else:  # no parameter attribute statement
                            self.modelParams[paramName] = paramValue
                tempLine = ''

    @property
    def pathObj(self):
        return self._pathObj

    @pathObj.setter
    def pathObj(self, value:pathlib.Path):
        assert isinstance(value,pathlib.Path)
        self._pathObj = value

    @property
    def pinOrder(self):
        return self._pinOrder

    @pinOrder.setter
    def pinOrder(self, value:str):
        assert isinstance(value,str)
        self._pinOrder = value

    @property
    def netlistLine(self):
        self._pinOrder = ','.join(self.pins)
        print(f'pinOrder : {self.pinOrder}')
        instParamString = ' '.join(
            [f'[@{key}:{key}=%:{key}={item}]' for key, item in
             self.instanceParams.items()])
        self._netlistLine = f'Y{self._vaModule} [@instName] [@pinList]  ' \
                            f'{self._vaModule}Model {instParamString}'
        return self._netlistLine

    @netlistLine.setter
    def netListLine(self, value:str):
        assert isinstance(value,str)
        self._netlistLine = value

    @property
    def vaModule(self):
        return self._vaModule




#   “Commons Clause” License Condition v1.0
#  #
#   The Software is provided to you by the Licensor under the License, as defined
#   below, subject to the following condition.
#  #
#   Without limiting other conditions in the License, the grant of rights under the
#   License will not include, and the License does not grant to you, the right to
#   Sell the Software.
#  #
#   For purposes of the foregoing, “Sell” means practicing any or all of the rights
#   granted to you under the License to provide to third parties, for a fee or other
#   consideration (including without limitation fees for hosting or consulting/
#   support services related to the Software), a product or service whose value
#   derives, entirely or substantially, from the functionality of the Software. Any
#   license notice or attribution required by the License must also include this
#   Commons Clause License Condition notice.
#  #
#   Software: Revolution EDA
#   License: Mozilla Public License 2.0
#   Licensor: Revolution Semiconductor (Registered in the Netherlands)

# properties dialogues for various symbol items

import pathlib

from PySide6.QtGui import (QFontDatabase, )
from PySide6.QtWidgets import (QDialog, QVBoxLayout, QHBoxLayout, QFormLayout,
                               QDialogButtonBox, QLineEdit, QLabel, QComboBox,
                               QGroupBox, QRadioButton, QGridLayout, QTextEdit,
                               QMenu)

import revedaEditor.common.net as net
import revedaEditor.common.shape as shp
import revedaEditor.gui.editFunctions as edf


class rectPropertyDialog(QDialog):
    '''
    Property dialog for symbol rectangles.
    '''
    def __init__(self, parent, rectItem: shp.rectangle):
        super().__init__(parent)
        self.parent = parent
        self.rectItem = rectItem
        self.location = self.rectItem.scenePos().toTuple()
        self.coords = self.rectItem.rect.getRect()

        self.setWindowTitle("Rectangle Properties")
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.mainLayout = QVBoxLayout()
        self.fLayout = QFormLayout()
        self.fLayout.setContentsMargins(10, 10, 10, 10)
        self.rectWidthLine = QLineEdit()
        self.rectWidthLine.setText(str(self.coords[2]))
        self.fLayout.addRow(QLabel("Width:"), self.rectWidthLine)
        self.rectHeightLine = QLineEdit()
        self.rectHeightLine.setText(str(self.coords[3]))
        self.fLayout.addRow(QLabel("Height:"), self.rectHeightLine)
        self.rectLeftLine = QLineEdit()
        self.rectLeftLine.setText(
            str(self.rectItem.start.toTuple()[0] + self.location[0]))
        self.fLayout.addRow(QLabel("X Origin:"), self.rectLeftLine)
        self.rectTopLine = QLineEdit()
        self.rectTopLine.setText(
            str(self.rectItem.start.toTuple()[1] + self.location[1]))
        self.fLayout.addRow(QLabel("Y Origin:"), self.rectTopLine)
        self.mainLayout.addLayout(self.fLayout)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()


class circlePropertyDialog(QDialog):
    def __init__(self, parent, circleItem: shp.circle):
        super().__init__(parent)
        self.parent = parent
        self.circleItem = circleItem
        self.location = self.circleItem.scenePos().toTuple()
        self.centre = self.circleItem.mapToScene(
            self.circleItem.centre).toTuple()
        self.radius = self.circleItem.radius

        self.setWindowTitle("Circle Properties")
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.mainLayout = QVBoxLayout()
        self.fLayout = QFormLayout()
        self.fLayout.setContentsMargins(10, 10, 10, 10)
        self.centerXEdit = edf.shortLineEdit()
        self.centerXEdit.setText(str(self.centre[0]))
        self.fLayout.addRow(QLabel("center x-coord:"), self.centerXEdit)
        self.centerYEdit = edf.shortLineEdit()
        self.centerYEdit.setText(str(self.centre[1]))
        self.fLayout.addRow(QLabel("center y-coord:"), self.centerYEdit)
        self.radiusEdit = edf.shortLineEdit()
        self.radiusEdit.setText(str(self.radius))
        self.fLayout.addRow(QLabel("radius:"), self.radiusEdit)
        self.mainLayout.addLayout(self.fLayout)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()

class arcPropertyDialog(QDialog):
    def __init__(self, parent, arcItem: shp.arc):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Arc Properties")
        self.arcItem = arcItem
        self.location = self.arcItem.scenePos().toTuple()
        # self.arcType = self.arcItem.arcType
        # self.arcTypeCombo = QComboBox()
        # self.arcTypeCombo.addItems(shp.arc.arcTypes)
        # self.arcTypeCombo.setCurrentText(self.arcType)

        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.mainLayout = QVBoxLayout()
        self.fLayout = QFormLayout()
        self.fLayout.setContentsMargins(10, 10, 10, 10)
        # self.mainLayout.addWidget(self.arcTypeCombo)
        self.startXEdit = edf.shortLineEdit()
        self.startXEdit.setText(
            str(self.arcItem.start.toTuple()[0] + self.location[0]))
        self.fLayout.addRow(QLabel("X Origin:"), self.startXEdit)
        self.startYEdit = edf.shortLineEdit()
        self.startYEdit.setText(
            str(self.arcItem.start.toTuple()[1] + self.location[1]))
        self.fLayout.addRow(QLabel("Y Origin:"), self.startYEdit)
        self.widthEdit = edf.shortLineEdit()
        self.widthEdit.setText(str(self.arcItem.width))
        self.fLayout.addRow(QLabel("Width:"), self.widthEdit)
        self.heightEdit = edf.shortLineEdit()
        self.heightEdit.setText(str(self.arcItem.height))
        self.fLayout.addRow(QLabel("Height:"), self.heightEdit)
        self.mainLayout.addLayout(self.fLayout)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()

class linePropertyDialog(QDialog):
    def __init__(self, parent, lineItem: shp.line):
        super().__init__(parent)
        self.parent = parent
        self.lineItem = lineItem
        self.location = self.lineItem.scenePos().toTuple()

        self.setWindowTitle("Line Properties")
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.mainLayout = QVBoxLayout()
        self.fLayout = QFormLayout()
        self.fLayout.setContentsMargins(10, 10, 10, 10)
        self.startXLine = QLineEdit()
        self.startXLine.setText(
            str(self.lineItem.start.toTuple()[0] + self.location[0]))
        self.fLayout.addRow(QLabel("Start (X):"), self.startXLine)
        self.startYLine = QLineEdit()
        self.startYLine.setText(
            str(self.lineItem.start.toTuple()[1] + self.location[1]))
        self.fLayout.addRow(QLabel("Start (Y):"), self.startYLine)
        self.endXLine = QLineEdit()
        self.endXLine.setText(
            str(self.lineItem.end.toTuple()[0] + self.location[0]))
        self.fLayout.addRow(QLabel("End (X):"), self.endXLine)
        self.endYLine = QLineEdit()
        self.endYLine.setText(
            str(self.lineItem.end.toTuple()[1] + self.location[1]))
        self.fLayout.addRow(QLabel("End (Y):"), self.endYLine)
        self.mainLayout.addLayout(self.fLayout)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()


class createPinDialog(QDialog):
    def __init__(self, parent) -> None:
        super().__init__(parent)
        self.setWindowTitle("Create Pin")
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.mainLayout = QVBoxLayout()
        self.fLayout = QFormLayout()
        self.pinName = QLineEdit()
        self.pinName.setPlaceholderText("Pin Name")
        self.pinName.setToolTip("Enter pin name")
        self.fLayout.addRow(QLabel("Pin Name"), self.pinName)
        self.pinDir = QComboBox()
        self.pinDir.addItems(shp.pin.pinDirs)
        self.pinDir.setToolTip("Select pin direction")
        self.fLayout.addRow(QLabel("Pin Direction"), self.pinDir)
        self.pinType = QComboBox()
        self.pinType.addItems(shp.pin.pinTypes)
        self.pinType.setToolTip("Select pin type")
        self.fLayout.addRow(QLabel("Pin Type"), self.pinType)
        self.mainLayout.addLayout(self.fLayout)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()


class pinPropertyDialog(createPinDialog):
    def __init__(self, parent, pinItem: shp.pin):
        super().__init__(parent)
        self.parent = parent
        self.pinItem = pinItem
        self.location = self.pinItem.scenePos().toTuple()

        self.setWindowTitle("Pin Properties")
        self.pinName.setText(str(pinItem.pinName))
        self.pinType.setCurrentText(pinItem.pinType)
        self.pinDir.setCurrentText(pinItem.pinDir)
        self.pinXLine = QLineEdit()
        self.pinXLine.setText(str(self.pinItem.start.x() + self.location[0]))
        self.pinXLine.setToolTip("X Coordinate")
        self.fLayout.addRow(QLabel("X:"), self.pinXLine)
        self.pinYLine = QLineEdit()
        self.pinYLine.setText(str(self.pinItem.start.y() + self.location[1]))
        self.pinYLine.setToolTip("Y Coordinate")
        self.fLayout.addRow(QLabel("Y:"), self.pinYLine)


class createSymbolLabelDialog(QDialog):
    def __init__(self, parent) -> None:
        super().__init__(parent)
        self.setWindowTitle("Create Label")
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.mainLayout = QVBoxLayout()
        self.fLayout = QFormLayout()
        self.labelDefinition = QLineEdit()
        self.labelDefinition.setPlaceholderText("Label Definition")
        self.labelDefinition.setToolTip("Enter label Definition")
        self.fLayout.addRow(QLabel("Label Definition"), self.labelDefinition)
        self.labelHeightEdit = QLineEdit()
        self.labelHeightEdit.setPlaceholderText("Label Height")
        self.labelHeightEdit.setToolTip("Enter label Height")
        self.fLayout.addRow(QLabel("Label Height"), self.labelHeightEdit)
        self.labelAlignCombo = QComboBox()
        self.labelAlignCombo.addItems(shp.label.labelAlignments)
        self.fLayout.addRow(QLabel("Label Alignment"), self.labelAlignCombo)
        self.labelOrientCombo = QComboBox()
        self.labelOrientCombo.addItems(shp.label.labelOrients)
        self.fLayout.addRow(QLabel("Label Orientation"), self.labelOrientCombo)
        self.labelUseCombo = QComboBox()
        self.labelUseCombo.addItems(shp.label.labelUses)
        self.fLayout.addRow(QLabel("Label Use"), self.labelUseCombo)
        self.labelVisiCombo = QComboBox()
        self.labelVisiCombo.addItems(["Yes", "No"])
        self.fLayout.addRow(QLabel("Label Visible"),self.labelVisiCombo)
        self.mainLayout.addLayout(self.fLayout)
        self.labelTypeGroup = QGroupBox("Label Type")
        self.labelTypeLayout = QHBoxLayout()
        self.normalType = QRadioButton(shp.label.labelTypes[0])
        self.normalType.setChecked(True)
        self.NLPType = QRadioButton(shp.label.labelTypes[1])
        self.pyLType = QRadioButton(shp.label.labelTypes[2])
        self.labelTypeLayout.addWidget(self.normalType)
        self.labelTypeLayout.addWidget(self.NLPType)
        self.labelTypeLayout.addWidget(self.pyLType)
        self.labelTypeGroup.setLayout(self.labelTypeLayout)
        self.mainLayout.addWidget(self.labelTypeGroup)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()


class labelPropertyDialog(createSymbolLabelDialog):
    def __init__(self, parent, labelItem: shp.label):
        assert isinstance(labelItem, shp.label)
        super().__init__(parent)
        self.parent = parent
        self.labelItem = labelItem
        self.location = self.labelItem.scenePos().toTuple()

        self.setWindowTitle("Label Properties")
        self.labelDefinition.setText(str(labelItem.labelDefinition))
        self.labelHeightEdit.setText(str(labelItem.labelHeight))
        self.labelAlignCombo.setCurrentText(labelItem.labelAlign)
        self.labelOrientCombo.setCurrentText(labelItem.labelOrient)
        self.labelUseCombo.setCurrentText(labelItem.labelUse)
        if labelItem.labelVisible:
            self.labelVisiCombo.setCurrentText("Yes")
        else:
            self.labelVisiCombo.setCurrentText("No")
        if self.labelItem.labelType == "Normal":
            self.normalType.setChecked(True)
        elif self.labelItem.labelType == "NLPLabel":
            self.NLPType.setChecked(True)
        elif self.labelItem.labelType == "PyLabel":
            self.pyLType.setChecked(True)
        self.labelXLine = QLineEdit()
        self.labelXLine.setText(
            str(self.labelItem.start.x() + self.location[0]))
        self.labelXLine.setToolTip("X Coordinate")
        self.fLayout.addRow(QLabel("X:"), self.labelXLine)
        self.labelYLine = QLineEdit()
        self.labelYLine.setText(
            str(self.labelItem.start.y() + self.location[1]))
        self.labelYLine.setToolTip("Y Coordinate")
        self.fLayout.addRow(QLabel("Y:"), self.labelYLine)


class symbolLabelsDialogue(QDialog):
    """
    Dialog for changing symbol labels and attributes. Symbol properties... menu item.
    """

    def __init__(self, parent, items: list, attributes: list):
        super().__init__(parent)
        self.parent = parent
        self.items = items
        self.attributes = attributes
        self.setWindowTitle("Symbol Labels")
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.mainLayout = QVBoxLayout()
        self.symbolPropsLayout = QGridLayout()
        self.symbolLabelsLayout = QGridLayout()
        self.symbolAttrsMethod()
        # Symbol Labels Layout
        self.symbolLabelsMethod()
        labelsGroup = QGroupBox("Symbol Labels")
        labelsGroup.setLayout(self.symbolLabelsLayout)
        self.mainLayout.addWidget(labelsGroup)
        self.mainLayout.addStretch(1)
        # self.mainLayout.addLayout(self.symbolLabelsLayout)
        propsGroup = QGroupBox("Symbol Attributes")
        propsGroup.setLayout(self.symbolPropsLayout)
        self.mainLayout.addWidget(propsGroup)
        self.mainLayout.addStretch(1)
        # self.mainLayout.addLayout(self.symbolPropsLayout)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

    def symbolAttrsMethod(self):
        self.attributeNameList = []
        # self.attributeTypeList = []
        self.attributeDefList = []
        # Symbol Properties
        self.symbolPropsLayout.addWidget(QLabel("Attribute Name"), 0, 0)
        # self.symbolPropsLayout.addWidget(QLabel("Type"), 0, 1)
        self.symbolPropsLayout.addWidget(QLabel("Definition"), 0, 1)
        i = 0
        for item in self.attributes:
            self.attributeNameList.append(edf.longLineEdit())
            self.attributeNameList[i].setText(item.name)
            self.symbolPropsLayout.addWidget(self.attributeNameList[i], i + 1,
                0)
            self.attributeDefList.append(edf.longLineEdit())
            self.attributeDefList[i].setText(item.definition)
            self.symbolPropsLayout.addWidget(self.attributeDefList[i], i + 1, 1)
            i += 1

        self.attributeNameList.append(edf.longLineEdit())
        self.attributeDefList.append(edf.longLineEdit())
        self.symbolPropsLayout.addWidget(self.attributeNameList[-1], i + 2, 0)
        self.symbolPropsLayout.addWidget(self.attributeDefList[-1], i + 2, 1)
        self.attributeNameList[-1].setPlaceholderText("Enter Attribute Name")
        self.attributeDefList[-1].setToolTip("Enter Attribute Definition")
        self.attributeDefList[-1].editingFinished.connect(
            lambda: self.updateAttributeDef(i + 1))

    def updateAttributeDef(self, i):
        i += 1
        self.attributeNameList.append(edf.longLineEdit())
        self.attributeDefList.append(edf.longLineEdit())
        self.symbolPropsLayout.addWidget(self.attributeNameList[-1], i + 1, 0)
        self.symbolPropsLayout.addWidget(self.attributeDefList[-1], i + 1, 1)
        self.attributeNameList[-1].setPlaceholderText("Enter Attribute Name")
        self.attributeDefList[-1].setToolTip("Enter Attribute Definition")
        self.attributeDefList[-1].editingFinished.connect(
            lambda: self.updateAttributeDef(i))

    def symbolLabelsMethod(self):
        self.labelDefinitionList = []
        self.labelHeightList = []
        self.labelAlignmentList = []
        self.labelOrientationList = []
        self.labelUseList = []
        self.labelTypeList = []
        self.labelItemList = []
        i = 0
        self.symbolLabelsLayout.addWidget(edf.boldLabel("Definition"), 0, 0)
        self.symbolLabelsLayout.addWidget(edf.boldLabel("Height"), 0, 1)
        self.symbolLabelsLayout.addWidget(edf.boldLabel("Alignment"), 0, 2)
        self.symbolLabelsLayout.addWidget(edf.boldLabel("Orientation"), 0, 3)
        self.symbolLabelsLayout.addWidget(edf.boldLabel("Use"), 0, 4)
        self.symbolLabelsLayout.addWidget(edf.boldLabel("Type"), 0, 5)
        for item in self.items:
            if type(item) == shp.label:
                i += 1
                self.labelItemList.append(item)
                self.labelDefinitionList.append(edf.longLineEdit())
                self.labelDefinitionList[-1].setText(item.labelDefinition)
                self.labelDefinitionList[-1].setReadOnly(True)
                self.symbolLabelsLayout.addWidget(
                    self.labelDefinitionList[i - 1], i, 0)
                self.labelHeightList.append(edf.shortLineEdit())
                self.labelHeightList[-1].setText(str(item.labelHeight))
                self.symbolLabelsLayout.addWidget(self.labelHeightList[i - 1],
                    i, 1)
                self.labelAlignmentList.append(QComboBox())
                self.labelAlignmentList[-1].addItems(shp.label.labelAlignments)
                self.labelAlignmentList[-1].setCurrentText(item.labelAlign)
                self.symbolLabelsLayout.addWidget(self.labelAlignmentList[-1],
                    i, 2)
                self.labelOrientationList.append(QComboBox())
                self.labelOrientationList[-1].addItems(shp.label.labelOrients)
                self.labelOrientationList[-1].setCurrentText(item.labelOrient)
                self.symbolLabelsLayout.addWidget(self.labelOrientationList[-1],
                    i, 3)
                self.labelUseList.append(QComboBox())
                self.labelUseList[-1].addItems(shp.label.labelUses)
                self.labelUseList[-1].setCurrentText(item.labelUse)
                self.symbolLabelsLayout.addWidget(self.labelUseList[-1], i, 4)
                self.labelTypeList.append(QComboBox())
                self.labelTypeList[-1].addItems(shp.label.labelTypes)
                self.labelTypeList[-1].setCurrentText(item.labelType)
                self.symbolLabelsLayout.addWidget(self.labelTypeList[-1], i, 5)
        if i == 0:  # no labels to edit
            self.symbolLabelsLayout.addWidget(QLabel("No symbol labels found."),
                1, 0)


class instanceProperties(QDialog):
    def __init__(self, parent, instance: shp.symbolShape = None):
        # assert isinstance(instance, shp.symbolShape)
        super().__init__(parent)
        self.parent = parent
        self.instance = instance
        self.initUI()

    def initUI(self):
        self.setWindowTitle("Instance Properties")
        self.mainLayout = QVBoxLayout()
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        formLayout = QFormLayout()
        self.libNameEdit = edf.longLineEdit()
        self.libNameEdit.setReadOnly(True)
        self.libNameEdit.setText(self.instance.libraryName)
        self.libNameEdit.setToolTip("Library Name (Read Only)")
        formLayout.addRow(edf.boldLabel("Library Name", self), self.libNameEdit)
        self.cellNameEdit = edf.longLineEdit()
        self.cellNameEdit.setText(self.instance.cellName)
        self.cellNameEdit.setReadOnly(True)
        self.cellNameEdit.setToolTip("Cell Name (Read Only)")
        formLayout.addRow(edf.boldLabel("Cell Name", self), self.cellNameEdit)
        self.viewNameEdit = edf.longLineEdit()
        self.viewNameEdit.setText(self.instance.viewName)
        self.viewNameEdit.setReadOnly(True)
        self.viewNameEdit.setToolTip("View Name (Read Only)")
        formLayout.addRow(edf.boldLabel("View Name", self), self.viewNameEdit)
        self.instNameEdit = edf.longLineEdit()
        self.instNameEdit.setText(self.instance.instanceName)
        self.instNameEdit.setToolTip("Instance Name")
        formLayout.addRow(edf.boldLabel("Instance Name", self), self.instNameEdit)
        location = (self.instance.scenePos() - self.instance.scene().origin).toTuple()
        self.xLocationEdit = edf.shortLineEdit()
        self.xLocationEdit.setText(str(location[0]))
        formLayout.addRow(edf.boldLabel("x location", self), self.xLocationEdit)
        self.yLocationEdit = edf.shortLineEdit()
        self.yLocationEdit.setText(str(location[1]))
        formLayout.addRow(edf.boldLabel("y location", self), self.yLocationEdit)
        self.angleEdit = edf.longLineEdit()
        self.angleEdit.setText(str(self.instance.angle))
        formLayout.addRow(edf.boldLabel("Angle", self), self.angleEdit)
        formLayout.setVerticalSpacing(10)
        self.instanceLabelsLayout = QGridLayout()
        self.instanceLabelsLayout.setColumnMinimumWidth(0, 100)
        self.instanceLabelsLayout.setColumnMinimumWidth(1, 200)
        self.instanceLabelsLayout.setColumnMinimumWidth(2, 100)
        self.instanceLabelsLayout.setColumnStretch(0, 0)
        self.instanceLabelsLayout.setColumnStretch(1, 1)
        self.instanceLabelsLayout.setColumnStretch(2, 0)
        row_index = 0
        for label in self.instance.labels.values():
            if label.labelDefinition not in shp.label.predefinedLabels:
                self.instanceLabelsLayout.addWidget(
                    edf.boldLabel(label.labelName, self), row_index, 0)
                labelValueEdit = edf.longLineEdit()
                labelValueEdit.setText(str(label.labelValue))
                self.instanceLabelsLayout.addWidget(labelValueEdit, row_index,
                    1)
                visibleCombo = QComboBox(self)
                visibleCombo.setInsertPolicy(QComboBox.NoInsert)
                visibleCombo.addItems(["True", "False"])
                if label.labelVisible:
                    visibleCombo.setCurrentIndex(0)
                else:
                    visibleCombo.setCurrentIndex(1)
                self.instanceLabelsLayout.addWidget(visibleCombo,row_index,2)
                row_index += 1

        instanceAttributesLayout = QGridLayout()
        instanceAttributesLayout.setColumnMinimumWidth(0, 100)
        instanceAttributesLayout.setColumnMinimumWidth(1, 200)
        instanceAttributesLayout.setColumnStretch(0, 0)
        instanceAttributesLayout.setColumnStretch(1, 1)
        # now list instance attributes
        for counter, name in enumerate(self.instance.attr.keys()):
            instanceAttributesLayout.addWidget(edf.boldLabel(name, self), counter,
                0)
            labelType = edf.longLineEdit()
            labelType.setReadOnly(True)
            labelNameEdit = edf.longLineEdit()
            labelNameEdit.setText(self.instance.attr.get(name))
            labelNameEdit.setToolTip(f"{name} attribute (Read Only)")
            instanceAttributesLayout.addWidget(labelNameEdit, counter, 1)

        self.mainLayout.addLayout(formLayout)
        self.mainLayout.addWidget(edf.boldLabel("Instance Labels", self))
        self.mainLayout.addLayout(self.instanceLabelsLayout)
        self.mainLayout.addSpacing(40)
        self.mainLayout.addWidget(edf.boldLabel("Instance Attributes", self))
        self.mainLayout.addLayout(instanceAttributesLayout)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)


class netProperties(QDialog):
    def __init__(self, parent, net: net.schematicNet = None):
        # assert isinstance(instance, shp.symbolShape)
        super().__init__(parent)
        self.parent = parent
        self.net = net
        self.initUI()

    def initUI(self):
        self.setWindowTitle("Net Properties")
        self.mainLayout = QVBoxLayout()
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        formBox = QGroupBox('Net Properties')
        formLayout = QFormLayout()
        self.netNameEdit = edf.longLineEdit()
        self.netNameEdit.setText(self.net.name)
        formLayout.addRow(edf.boldLabel("Net Name", self), self.netNameEdit)
        formBox.setLayout(formLayout)
        self.mainLayout.addWidget(formBox)
        self.mainLayout.addSpacing(40)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)


class createSchematicPinDialog(createPinDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.setWindowTitle("Create Schematic Pin")


class schematicPinPropertiesDialog(createPinDialog):
    def __init__(self, parent, item):
        super().__init__(parent)
        self.setWindowTitle(f"{item.pinName} - Pin Properties")
        self.pinName.setText(item.pinName)
        self.pinDir.setCurrentText(item.pinDir)
        self.pinType.setCurrentText(item.pinType)

class symbolNameDialog(QDialog):
    def __init__(self, cellPath:pathlib.Path, cellName: str, parent):
        super().__init__(parent)
        self.cellPath = cellPath
        self.cellName = cellName
        self.symbolViewNames = [view.stem for view in cellPath.iterdir() if 'symbol' in
                                view.stem]
        self.initUI()
    def initUI(self):
        self.setWindowTitle('Create a symbol?')
        self.mainLayout = QVBoxLayout()
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        formLayout = QFormLayout()
        formLayout.addRow(edf.boldLabel('Library Name'), QLabel(
            self.cellPath.parent.stem) )
        formLayout.addRow(edf.boldLabel('Cell Name'), QLabel(self.cellPath.stem))
        self.symbolViewsCB = QComboBox()
        self.symbolViewsCB.addItems(self.symbolViewNames)
        self.symbolViewsCB.setEditable(True)
        formLayout.addRow(edf.boldLabel('Symbol View Name:'), self.symbolViewsCB)
        self.mainLayout.addLayout(formLayout)
        self.mainLayout.addSpacing(40)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)

class symbolCreateDialog(QDialog):
    def __init__(self, parent, inputPinNames: list, outputPinNames: list,
                 inoutPinNames: list):
        super().__init__(parent)
        self.parent = parent
        self.inputPinNames = inputPinNames
        self.outputPinNames = outputPinNames
        self.inoutPinNames = inoutPinNames
        self.setWindowTitle("Create Symbol")
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.mainLayout = QVBoxLayout()

        self.fLayout = QFormLayout()
        self.topPinsEdit = edf.longLineEdit()
        self.topPinsEdit.setText(', '.join(self.inoutPinNames))
        self.topPinsEdit.setToolTip("Enter top pins")
        self.fLayout.addRow(edf.boldLabel("Top Pins:"), self.topPinsEdit)
        self.leftPinsEdit = edf.longLineEdit()
        self.leftPinsEdit.setText(', '.join(self.inputPinNames))
        self.leftPinsEdit.setToolTip("Enter left pins")
        self.fLayout.addRow(edf.boldLabel("Left Pins:"), self.leftPinsEdit)
        self.bottomPinsEdit = edf.longLineEdit()
        self.bottomPinsEdit.setToolTip("Enter bottom pins")
        self.fLayout.addRow(edf.boldLabel("Bottom Pins:"), self.bottomPinsEdit)
        self.rightPinsEdit = edf.longLineEdit()
        self.rightPinsEdit.setText(', '.join(self.outputPinNames))
        self.rightPinsEdit.setToolTip("Enter right pins")
        self.fLayout.addRow(edf.boldLabel("Right Pins:"), self.rightPinsEdit)
        self.mainLayout.addLayout(self.fLayout)
        self.mainLayout.addSpacing(20)
        self.geomLayout = QFormLayout()
        self.stubLengthEdit = QLineEdit()
        self.stubLengthEdit.setText('60')
        self.stubLengthEdit.setToolTip('Enter stub lengths')
        self.geomLayout.addRow(edf.boldLabel("Stub Length:"), self.stubLengthEdit)
        self.pinDistanceEdit = QLineEdit()
        self.pinDistanceEdit.setText('80')
        self.pinDistanceEdit.setToolTip('Enter pin spacing')
        self.geomLayout.addRow(edf.boldLabel("Pin spacing:"), self.pinDistanceEdit)
        self.mainLayout.addLayout(self.geomLayout)
        self.mainLayout.addSpacing(40)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.mainLayout.addWidget(self.buttonBox)
        self.setLayout(self.mainLayout)
        self.show()


class noteTextEdit(QDialog):
    '''
    Set text properties.
    '''
    def __init__(self,parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Edit Text")
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        mainLayout = QVBoxLayout()
        self.plainTextEdit = QTextEdit()
        mainLayout.addWidget(self.plainTextEdit)
        fontFamilies = QFontDatabase.families(QFontDatabase.Latin)
        fixedFamilies = [family for family in fontFamilies if
                         QFontDatabase.isFixedPitch(family)]
        formLayout = QFormLayout()
        self.familyCB = QComboBox()
        self.familyCB.addItems(fixedFamilies)
        self.familyCB.currentTextChanged.connect(self.familyFontStyles)
        formLayout.addRow(edf.boldLabel('Font Name'),self.familyCB)
        self.fontStyleCB = QComboBox()
        self.fontStyles = QFontDatabase.styles(fixedFamilies[0])
        self.fontStyleCB.addItems(self.fontStyles)
        self.fontStyleCB.currentTextChanged.connect(self.styleFontSizes)
        formLayout.addRow(edf.boldLabel('Font Style'), self.fontStyleCB)
        self.fontsizeCB = QComboBox()
        self.fontSizes = [str(size) for size in QFontDatabase.pointSizes(fixedFamilies[0],
                                                          self.fontStyles[0])]
        self.fontsizeCB.addItems(self.fontSizes)
        formLayout.addRow(edf.boldLabel('Font Size'),self.fontsizeCB)
        self.textAlignmCB = QComboBox()
        self.textAlignmCB.addItems(shp.text.textAlignments)
        formLayout.addRow(edf.boldLabel('Text Alignment'),self.textAlignmCB)
        self.textOrientCB = QComboBox()
        self.textOrientCB.addItems(shp.text.textOrients)
        formLayout.addRow(edf.boldLabel('Text Orientation'), self.textOrientCB)
        mainLayout.addLayout(formLayout)
        mainLayout.addSpacing(40)
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        mainLayout.addWidget(self.buttonBox)
        self.setLayout(mainLayout)
        self.show()

    def familyFontStyles(self,s):
        self.fontStyleCB.clear()
        self.fontStyles = QFontDatabase.styles(self.familyCB.currentText())
        self.fontStyleCB.addItems(self.fontStyles)

    def styleFontSizes(self,s):
        self.fontsizeCB.clear()
        selectedFamily = self.familyCB.currentText()
        selectedStyle = self.fontStyleCB.currentText()
        self.fontSizes = [str(size) for size in QFontDatabase.pointSizes(
            selectedFamily, selectedStyle)]
        self.fontsizeCB.addItems(self.fontSizes)

class noteTextEditProperties(noteTextEdit):
    def __init__(self,parent, note:shp.text):
        super().__init__(parent)
        self.note = note
        self.plainTextEdit.setText(self.note.textContent)
        self.familyCB.setCurrentText(self.note.fontFamily)
        self.fontStyleCB.setCurrentText(self.note.fontStyle)
        self.fontsizeCB.setCurrentText(self.note.textHeight)
        self.textAlignmCB.setCurrentText(self.note.textAlignment)
        self.textOrientCB.setCurrentText(self.note.textOrient)


class displayConfigDialog(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Display Options")
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        self.vLayout = QVBoxLayout()
        fLayout = QFormLayout()
        self.majorGridEntry = QLineEdit()
        fLayout.addRow("Major Grid:", self.majorGridEntry)
        self.snapDistanceEntry = QLineEdit()
        fLayout.addRow("Snap Distance", self.snapDistanceEntry)

        gridTypeGroup = QGroupBox("Grid Type")
        gridTypeLayout = QHBoxLayout()
        self.dotType = QRadioButton('Dot Grid')
        self.dotType.setChecked(True)
        self.lineType = QRadioButton('Line Grid')
        self.noType = QRadioButton('No Grid')
        gridTypeLayout.addWidget(self.dotType)
        gridTypeLayout.addWidget(self.lineType)
        gridTypeLayout.addWidget(self.noType)
        gridTypeGroup.setLayout(gridTypeLayout)

        self.vLayout.addLayout(fLayout)
        self.vLayout.addWidget(gridTypeGroup)
        self.vLayout.addStretch(1)

        self.vLayout.addWidget(self.buttonBox)
        self.setLayout(self.vLayout)
        self.show()

class selectConfigDialogue(QDialog):
    def __init__(self, parent):
        super().__init__(parent)
        self.parent = parent
        self.setWindowTitle("Selection Options")
        QBtn = QDialogButtonBox.Ok | QDialogButtonBox.Cancel
        self.buttonBox = QDialogButtonBox(QBtn)
        self.buttonBox.accepted.connect(self.accept)
        self.buttonBox.rejected.connect(self.reject)
        vLayout = QVBoxLayout()
        selectionTypeGroup = QGroupBox("Selection Type")
        selectionTypeLayout = QHBoxLayout()
        self.fullSelection = QRadioButton('Full')
        self.partialSelection = QRadioButton('Partial')
        selectionTypeLayout.addWidget(self.fullSelection)
        selectionTypeLayout.addWidget(self.partialSelection)
        selectionTypeGroup.setLayout(selectionTypeLayout)
        vLayout.addWidget(selectionTypeGroup)
        vLayout.addStretch(1)
        vLayout.addWidget(self.buttonBox)
        self.setLayout(vLayout)
        self.show()



#   “Commons Clause” License Condition v1.0
#  #
#   The Software is provided to you by the Licensor under the License, as defined
#   below, subject to the following condition.
#  #
#   Without limiting other conditions in the License, the grant of rights under the
#   License will not include, and the License does not grant to you, the right to
#   Sell the Software.
#  #
#   For purposes of the foregoing, “Sell” means practicing any or all of the rights
#   granted to you under the License to provide to third parties, for a fee or other
#   consideration (including without limitation fees for hosting or consulting/
#   support services related to the Software), a product or service whose value
#   derives, entirely or substantially, from the functionality of the Software. Any
#   license notice or attribution required by the License must also include this
#   Commons Clause License Condition notice.
#  #
#   Software: Revolution EDA
#   License: Mozilla Public License 2.0
#   Licensor: Revolution Semiconductor (Registered in the Netherlands)
